import argparse
import datetime
import functools
import json
import os
import pprint
import urllib
import requests
import sys
import time

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse


TIME_FORMAT = "%Y-%m-%dT%H:%M:%S"


class FailedToLogin(Exception):
    pass


class DeviceNotFound(Exception):
    pass


class RequestFailed(Exception):
    pass


class Device(object):
    def __init__(self, client, ident):
        self._client = client
        self.ident = ident
        self._devinfo = {}
        self._ports = {}

    def refresh(self, info=None):
        if info is None:
            info = self._client._find_device(ident=self.ident)
        self._devinfo = info

    def set_port(self, port):
        self._ports[port.ident] = port

    @property
    def ports(self):
        return self._ports

    @property
    def data(self):
        return self._devinfo


class Port(object):
    def __init__(self, client, ident):
        self._client = client
        self.ident = ident
        self._portinfo = {}

    def refresh(self, info=None):
        if info is None:
            sensors = self._client._get_sensors()
            info = self._client._find_sensor(sensors, self.ident)
        self._portinfo.update(info)

    def __repr__(self):
        try:
            data = '%s=%s' % (self.tag, self.value)
        except ValueError:
            data = '?'
        return '<Port %s %s>' % (self.ident, data)

    @property
    def value(self):
        if 'tag' not in self._portinfo:
            raise ValueError('Port has no value')
        return self._portinfo.get(self.tag)

    @property
    def tag(self):
        if 'tag' not in self._portinfo:
            raise ValueError('Port is not initialized')
        return self._portinfo['tag']

    @property
    def data(self):
        return self._portinfo

    @property
    def label(self):
        return self._portinfo['label']

    @property
    def model(self):
        return self._portinfo['model']

    @property
    def output(self):
        return self._portinfo.get('output')

    def control(self, state):
        self._client._control_port(self.ident, state)


def retries_login(fn):
    @functools.wraps(fn)
    def wrapper(self, *args, **kwargs):
        for i in (0, 1):
            try:
                return fn(self, *args, **kwargs)
            except RequestFailed:
                if i == 0:
                    self._login()
                else:
                    # Make sure we raise the original exception
                    # if we retried login already and still explode
                    raise
    return wrapper


class MFiClient(object):
    def __init__(self, host, username, password, port=None,
                 use_tls=True, verify=True):
        self._host = host
        self._port = port
        self._user = username
        self._pass = password
        self._stat_cache = None
        self._cookie = None
        self._session = requests.Session()
        self._verify = verify
        if use_tls:
            port = port or 6443
            self._baseurl = 'https://%s:%i' % (host, port)
        else:
            port = port or 6080
            self._baseurl = 'http://%s:%i' % (host, port)

        self._login()

    def _login(self):
        response = self._session.get(self._baseurl, verify=self._verify)

        data = {'username': self._user,
                'password': self._pass,
                'login': 'Login'}

        response = self._session.post('%s/login' % self._baseurl,
                                      data=data, verify=self._verify)
        if response.status_code == 200 and response.url.endswith('/manage'):
            return

        raise FailedToLogin('Server rejected login')

    @retries_login
    def _get_stat(self):
        response = self._session.get('%s/api/v1.0/stat/device' % self._baseurl,
                                     verify=self._verify)
        if response.status_code == 200:
            return response.json()['data']
        raise RequestFailed()

    @retries_login
    def _get_sensors(self):
        data = {'json': json.dumps({'hello': 2})}
        response = self._session.post(
            '%s/api/v1.0/list/sensors' % self._baseurl, data=data,
            verify=self._verify)
        if response.status_code == 200:
            return response.json()['data']
        raise RequestFailed()

    get_raw_sensors = _get_sensors
    get_raw_status = _get_stat

    @staticmethod
    def _find_sensor(sensors, ident):
        for sensor in sensors:
            if sensor['_id'] == ident:
                return sensor
        raise DeviceNotFound('No sensor %s' % ident)

    def get_devices(self):
        stat = self._get_stat()
        sensors = self._get_sensors()

        devices = []
        for devinfo in stat:
            device = Device(self, devinfo['_id'])
            for portinfo in devinfo['port_cfg']:
                if portinfo['_id'] == 'NONE':
                    continue
                sensorinfo = self._find_sensor(sensors, portinfo['_id'])

                port = Port(self, portinfo['_id'])
                port.refresh(portinfo)
                port.refresh(sensorinfo)
                device.set_port(port)
            devices.append(device)
        return devices

    def get_port(self, ident=None, label=None):
        for device in self.get_devices():
            for port in device.ports.values():
                if port.label == label or port.ident == ident:
                    return port
        return None

    def get_stat(self):
        if not self._stat_cache:
            self._stat_cache = self._get_stat()
        return self._stat_cache

    @retries_login
    def _control_port(self, ident, state, voltage=0):
        the_port = self._find_port(ident=ident)

        voltages = {
            'Output 5v': 5,
            'Output 12v': 12,
            'Output 24v': 24,
        }
        voltage = state and voltages.get(the_port['model'], 0) or 0

        data = {
            'sId': ident,
            'mac': the_port['mac'],
            'model': the_port['model'],
            'port': int(the_port['port']),
            'cmd': 'mfi-output',
            'val': int(state),
            'volt': voltage,
        }
        data = {'json': json.dumps(data)}
        response = self._session.post('%s/api/v1.0/cmd/devmgr' % self._baseurl,
                                      data=data, verify=self._verify)
        if response.status_code == 200:
            return response.text
        raise RequestFailed()

    def _find_port(self, ident=None, device_name=None):
        devices = self.get_stat()

        for dev in devices:
            for port in dev['port_cfg']:
                if port['_id'] == ident:
                    return port
                if port['label'] == device_name:
                    return port
        raise DeviceNotFound('No such device')

    def get_device_data(self, device, since=60):
        # NOTE: This is broken
        port = self._find_device(device)
        sensor = self._find_sensor(device)

        start = time.time() - since
        end = time.time()
        data = {
            'fmt': 'json',
            'ids': port['_id'],
            'tags': sensor['tag'],
            'indices': '1,2,3,4',
            'func': 'trend',
            'collection': 'null',
            'startTime': int(start) * 1000,
            'endTime': int(end) * 1000,
        }
        response = self._session.get(
            '%s/api/v1.0/data/m2mgeneric_by_id' % self._baseurl,
            params=data, verify=self._verify)
        return response.json()['data'][0]['%s.0' % sensor['tag']]


def get_auth_from_env():
    """Attempt to get mFi connection information from the environment.

    Supports either a combined variable called MFI formatted like:

        MFI="http://user:pass@192.168.1.1:6080/

    or individual ones like:

        MFI_HOST=192.168.1.1
        MFI_PORT=6080
        MFI_USER=foo
        MFI_PASS=pass

    :returns: A tuple like (host, port, user, pass, path)
    """

    combined = os.getenv('MFI')
    if combined:
        # http://user:pass@192.168.1.1:7080/
        result = urlparse.urlparse(combined)
        netloc = result.netloc
        if '@' in netloc:
            creds, netloc = netloc.split('@', 1)
            user, _pass = creds.split(':', 1)
        else:
            user = 'mfiadmin'
            _pass = 'password'
        if ':' in netloc:
            host, port = netloc.split(':', 1)
            port = int(port)
        else:
            host = netloc
            port = 6080
        path = result.path
        tls = combined.startswith('https://')
    else:
        host = os.getenv('MFI_HOST')
        port = int(os.getenv('MFI_PORT', 7080))
        user = os.getenv('MFI_USER')
        _pass = os.getenv('MFI_PASS')
        path = '/'
        tls = False
    return host, port, user, _pass, path, tls


def envclient():
    host, port, user, _pass, path = get_auth_from_env()
    return MFiClient(host, user, _pass, port=port)



import json
import os

from mficlient import client
from mficlient import fake_data


class FakeResponse(object):
    def __init__(self, code, data):
        self.status_code = code
        self._data = data

    def json(self):
        return {'data': self._data}

    @property
    def text(self):
        return self._data


class FakeSession(object):
    def __init__(self):
        self._status = json.loads(fake_data.FAKE_STATUS)
        self._sensors = json.loads(fake_data.FAKE_SENSORS)

    def get(self, url, headers=None, verify=True):
        if url.endswith('stat/device'):
            return FakeResponse(200, self._status)
        elif url.endswith('list/sensors'):
            return FakeResponse(200, self._sensors)
        else:
            raise Exception('Unsupported fake path %s' % url)

    def _do_device(self, data):
        cmd = json.loads(data['json'])
        updates = dict(cmd)
        ident = updates.pop('sId')
        for sensor in self._sensors:
            if sensor['_id'] == ident:
                sensor.update(updates)
            sensor['output'] = float(sensor['val'] > 0)
        return FakeResponse(200, self._sensors)

    def post(self, url, data=None, headers=None, verify=True):
        if url.endswith('list/sensors'):
            return FakeResponse(200, self._sensors)
        elif url.endswith('cmd/devmgr'):
            return self._do_device(data)


class FakeMFiClient(client.MFiClient):
    def __init__(self, *args, **kwargs):
        if not args:
            args = ('fakehost', 'fakeuser', 'fakepass')
        super(FakeMFiClient, self).__init__(*args, **kwargs)
        self._session = FakeSession()

    def _login(self):
        pass



from setuptools import setup

setup(name='mficlient',
      version='0.3.0',
      description='A remote control client for Ubiquiti\'s mFi system',
      author='Dan Smith',
      author_email='dsmith+mficlient@danplanet.com',
      url='http://github.org/kk7ds/mficlient',
      packages=['mficlient'],
      scripts=['mfi'],
      install_requires=['requests'],
      tests_require=['mock'],
)



import argparse
import json
import os
import pprint
import requests

from mficlient import client


class Application(object):
    def main(self):
        try:
            requests.packages.urllib3.disable_warnings()
        except:
            pass
        commands = [x[4:] for x in dir(self) if x.startswith('cmd')]
        command_list = os.linesep + os.linesep.join(commands)

        parser = argparse.ArgumentParser()
        parser.add_argument('command',
                            help='One of: %s' % command_list)
        parser.add_argument('--device',
                            help='Specific device')
        parser.add_argument('--property',
                            help='Show only this property of a device')
        parser.add_argument('--state',
                            help='State to set (on or off)')
        parser.add_argument('--every', type=int, default=0,
                            help='Repeat (interval in seconds)')
        parser.add_argument('--since', type=int, default=60,
                            metavar='SECS',
                            help='Show data since SECS seconds ago')
        parser.add_argument('--column-headers', default=False,
                            action='store_true',
                            help='Show CSV column headers')
        parser.add_argument('--json', default=False,
                            action='store_true',
                            help='Output raw JSON for the raw_ commands')
        parser.add_argument('--noverify', default=False,
                            action='store_true',
                            help='Do not verify server SSL certificate')
        args = parser.parse_args()

        if not hasattr(self, 'cmd_%s' % args.command):
            print('No such command `%s\'' % args.command)
            return 1

        host, port, user, _pass, path, tls = client.get_auth_from_env()
        self._client = client.MFiClient(host, user, _pass, port=port,
                                        use_tls=tls, verify=not args.noverify)

        while True:
            getattr(self, 'cmd_%s' % args.command)(args)
            if not args.every:
                break
            time.sleep(args.every)

    def cmd_dump_sensors(self, options):
        devices = self._client.get_devices()

        fmt = '%20s | %20s | %15s | %10s | %s'
        print(fmt % ('Model', 'Label', 'Tag', 'Value', 'Extra'))
        print('-' * 78)
        for device in devices:
            for port in device.ports.values():
                print(fmt % (port.model, port.label,
                             port.tag, port.value,
                             port.output))

    def cmd_raw_sensors(self, options):
        data = self._client.get_raw_sensors()
        if options.device:
            data = [x for x in data if x['label'] == options.device]
        if options.json:
            print(json.dumps(data))
        else:
            pprint.pprint(data)

    def cmd_raw_status(self, options):
        data = self._client.get_raw_status()
        if options.json:
            print(json.dumps(data))
        else:
            pprint.pprint(data)

    def cmd_control_device(self, options):
        if not options.device:
            print('Must specify a device')
            return
        if not options.state:
            print('Must specify a state')
            return
        port = self._client.get_port(label=options.device)
        port.control(options.state == 'on')

    def cmd_get_data(self, options):
        if not options.device:
            print('Must specify a device')
            return

        devices = self._client.get_devices()
        port = None
        for dev in devices:
            for port in dev.ports.values():
                if port.label == options.device:
                    break
        if port is None:
            print('No such port %s' % options.device)
            return

        headers = list(port.data.keys())
        if options.property:
            if options.property not in headers:
                print('Port has no property `%s`' % options.property)
                return
            headers = [options.property]

        data = [str(port.data[key]) for key in headers]

        if options.column_headers:
            print(','.join(headers))
        print(','.join(data))

    def cmd_sensors_csv(self, options):
        if not options.device:
            print('Must specify a device')
            return
        sensors = self._client.get_raw_sensors()
        keys = ['active_pwr', 'energy_sum', 'i_rms', 'v_rms',
                'label', 'model', 'output', 'output_val', 'pf',
                'port', 'tag', 'val', 'wattHours', 'wattHoursBase']
        the_sensor = None
        for sensor in sensors:
            if sensor['label'] == options.device:
                the_sensor = sensor
                break

        if not the_sensor:
            print('No such device `%s\'' % options.device)
            return

        rpt_time = datetime.datetime.fromtimestamp(
            sensor['rpt_time'] / 1000)
        try:
            wh_rpt_time = datetime.datetime.fromtimestamp(
                sensor['wh_rpt_time'] / 1000)
        except KeyError:
            wh_rpt_time = ''

        vals = [str(sensor.get(k, '')) for k in keys]
        vals.insert(0, rpt_time.strftime(TIME_FORMAT))
        vals.append(wh_rpt_time and wh_rpt_time.strftime(TIME_FORMAT) or '')
        keys.insert(0, 'time')
        keys.append('wh_rpt_time')

        if options.column_headers:
            print(','.join(keys))

        print(','.join(vals))
